import { writeUndiscriminatableMorphUnionMessage } from "../parse/string/ast.js";
import { domainOf } from "../utils/domains.js";
import { throwParseError } from "../utils/errors.js";
import { hasKey, isKeyOf, keysOf } from "../utils/generics.js";
import { isArray, objectKindOf } from "../utils/objectKinds.js";
import { Path } from "../utils/paths.js";
import { serializePrimitive } from "../utils/serialize.js";
import { branchIntersection, flattenBranch } from "./branch.js";
import { IntersectionState } from "./compose.js";
import { mappedKeys, propToNode } from "./rules/props.js";
export const flattenBranches = (branches, ctx)=>{
    const discriminants = calculateDiscriminants(branches, ctx);
    const indices = branches.map((_, i)=>i);
    return discriminate(branches, indices, discriminants, ctx);
};
const discriminate = (originalBranches, remainingIndices, discriminants, ctx)=>{
    if (remainingIndices.length === 1) {
        return flattenBranch(originalBranches[remainingIndices[0]], ctx);
    }
    const bestDiscriminant = findBestDiscriminant(remainingIndices, discriminants);
    if (!bestDiscriminant) {
        return [
            [
                "branches",
                remainingIndices.map((i)=>branchIncludesMorph(originalBranches[i], ctx.type.meta.scope) ? throwParseError(writeUndiscriminatableMorphUnionMessage(`${ctx.path}`)) : flattenBranch(originalBranches[i], ctx))
            ]
        ];
    }
    const cases = {};
    let caseKey;
    for(caseKey in bestDiscriminant.indexCases){
        const nextIndices = bestDiscriminant.indexCases[caseKey];
        cases[caseKey] = discriminate(originalBranches, nextIndices, discriminants, ctx);
    }
    return [
        [
            "switch",
            {
                path: bestDiscriminant.path,
                kind: bestDiscriminant.kind,
                cases
            }
        ]
    ];
};
const discriminantKinds = {
    domain: true,
    objectKind: true,
    value: true
};
const calculateDiscriminants = (branches, ctx)=>{
    const discriminants = {
        disjointsByPair: {},
        casesByDisjoint: {}
    };
    for(let lIndex = 0; lIndex < branches.length - 1; lIndex++){
        for(let rIndex = lIndex + 1; rIndex < branches.length; rIndex++){
            const pairKey = `${lIndex}/${rIndex}`;
            const pairDisjoints = [];
            discriminants.disjointsByPair[pairKey] = pairDisjoints;
            const intersectionState = new IntersectionState(ctx.type, "|");
            branchIntersection(branches[lIndex], branches[rIndex], intersectionState);
            for(const path in intersectionState.disjoints){
                if (path.includes(mappedKeys.index)) {
                    continue;
                }
                const { l , r , kind  } = intersectionState.disjoints[path];
                if (!isKeyOf(kind, discriminantKinds)) {
                    continue;
                }
                const lSerialized = serializeIfAllowed(kind, l);
                const rSerialized = serializeIfAllowed(kind, r);
                if (lSerialized === undefined || rSerialized === undefined) {
                    continue;
                }
                const qualifiedDisjoint = path === "/" ? kind : `${path}/${kind}`;
                pairDisjoints.push(qualifiedDisjoint);
                if (!discriminants.casesByDisjoint[qualifiedDisjoint]) {
                    discriminants.casesByDisjoint[qualifiedDisjoint] = {
                        [lSerialized]: [
                            lIndex
                        ],
                        [rSerialized]: [
                            rIndex
                        ]
                    };
                    continue;
                }
                const cases = discriminants.casesByDisjoint[qualifiedDisjoint];
                if (!hasKey(cases, lSerialized)) {
                    cases[lSerialized] = [
                        lIndex
                    ];
                } else if (!cases[lSerialized].includes(lIndex)) {
                    cases[lSerialized].push(lIndex);
                }
                if (!hasKey(cases, rSerialized)) {
                    cases[rSerialized] = [
                        rIndex
                    ];
                } else if (!cases[rSerialized].includes(rIndex)) {
                    cases[rSerialized].push(rIndex);
                }
            }
        }
    }
    return discriminants;
};
const parseQualifiedDisjoint = (qualifiedDisjoint)=>{
    const path = Path.fromString(qualifiedDisjoint);
    return [
        path,
        path.pop()
    ];
};
const findBestDiscriminant = (remainingIndices, discriminants)=>{
    let bestDiscriminant;
    for(let i = 0; i < remainingIndices.length - 1; i++){
        const lIndex = remainingIndices[i];
        for(let j = i + 1; j < remainingIndices.length; j++){
            const rIndex = remainingIndices[j];
            const candidates = discriminants.disjointsByPair[`${lIndex}/${rIndex}`];
            for (const qualifiedDisjoint of candidates){
                const indexCases = discriminants.casesByDisjoint[qualifiedDisjoint];
                const filteredCases = {};
                const defaultCases = [
                    ...remainingIndices
                ];
                let score = 0;
                let caseKey;
                for(caseKey in indexCases){
                    const filteredIndices = indexCases[caseKey].filter((i)=>{
                        const remainingIndex = remainingIndices.indexOf(i);
                        if (remainingIndex !== -1) {
                            delete defaultCases[remainingIndex];
                            return true;
                        }
                    });
                    if (filteredIndices.length === 0) {
                        continue;
                    }
                    filteredCases[caseKey] = filteredIndices;
                    score++;
                }
                const defaultCaseKeys = keysOf(defaultCases);
                if (defaultCaseKeys.length) {
                    filteredCases["default"] = defaultCaseKeys.map((k)=>parseInt(k));
                }
                if (!bestDiscriminant || score > bestDiscriminant.score) {
                    const [path, kind] = parseQualifiedDisjoint(qualifiedDisjoint);
                    bestDiscriminant = {
                        path,
                        kind,
                        indexCases: filteredCases,
                        score
                    };
                    if (score === remainingIndices.length) {
                        // if we find a candidate that discriminates all branches, return early
                        return bestDiscriminant;
                    }
                }
            }
        }
    }
    return bestDiscriminant;
};
export const serializeIfAllowed = (kind, data)=>kind === "value" ? serializeIfPrimitive(data) : `${data}`;
const serializeIfPrimitive = (data)=>{
    const domain = domainOf(data);
    return domain === "object" || domain === "symbol" ? undefined : serializePrimitive(data);
};
const caseSerializers = {
    value: (data)=>serializeIfPrimitive(data) ?? "default",
    objectKind: (data)=>objectKindOf(data) ?? "default",
    domain: domainOf
};
export const serializeCase = (kind, data)=>caseSerializers[kind](data);
const branchIncludesMorph = (branch, $)=>"morph" in branch ? true : "props" in branch ? Object.values(branch.props).some((prop)=>nodeIncludesMorph(propToNode(prop), $)) : false;
const nodeIncludesMorph = (node, $)=>typeof node === "string" ? $.resolve(node).meta.includesMorph : Object.values(node).some((predicate)=>predicate === true ? false : isArray(predicate) ? predicate.some((branch)=>branchIncludesMorph(branch, $)) : branchIncludesMorph(predicate, $));
