import { compileDisjointReasonsMessage } from "../parse/string/ast.js";
import { throwParseError } from "../utils/errors.js";
import { hasKey, hasKeys, keysOf } from "../utils/generics.js";
import { Path } from "../utils/paths.js";
import { anonymousDisjoint, composeKeyedIntersection, IntersectionState, isDisjoint, isEquality, throwUndefinedOperandsError } from "./compose.js";
import { flattenPredicate, isLiteralCondition, predicateIntersection, predicateUnion, resolutionExtendsDomain } from "./predicate.js";
import { mappedKeys } from "./rules/props.js";
export const nodeIntersection = (l, r, state)=>{
    state.domain = undefined;
    const lResolution = state.type.meta.scope.resolveNode(l);
    const rResolution = state.type.meta.scope.resolveNode(r);
    const result = resolutionIntersection(lResolution, rResolution, state);
    if (typeof result === "object" && !hasKeys(result)) {
        return hasKeys(state.disjoints) ? anonymousDisjoint() : state.addDisjoint("domain", keysOf(lResolution), keysOf(rResolution));
    }
    return result === lResolution ? l : result === rResolution ? r : result;
};
const resolutionIntersection = composeKeyedIntersection((domain, l, r, context)=>{
    if (l === undefined) {
        return r === undefined ? throwUndefinedOperandsError() : undefined;
    }
    if (r === undefined) {
        return undefined;
    }
    return predicateIntersection(domain, l, r, context);
}, {
    onEmpty: "omit"
});
export const intersection = (l, r, type)=>{
    const state = new IntersectionState(type, "&");
    const result = nodeIntersection(l, r, state);
    return isDisjoint(result) ? throwParseError(compileDisjointReasonsMessage(state.disjoints)) : isEquality(result) ? l : result;
};
export const union = (l, r, type)=>{
    const lResolution = type.meta.scope.resolveNode(l);
    const rResolution = type.meta.scope.resolveNode(r);
    const result = {};
    const domains = keysOf({
        ...lResolution,
        ...rResolution
    });
    for (const domain of domains){
        result[domain] = hasKey(lResolution, domain) ? hasKey(rResolution, domain) ? predicateUnion(domain, lResolution[domain], rResolution[domain], type) : lResolution[domain] : hasKey(rResolution, domain) ? rResolution[domain] : throwUndefinedOperandsError();
    }
    return result;
};
const hasImpliedDomain = (flatPredicate)=>flatPredicate[0] && (flatPredicate[0][0] === "value" || flatPredicate[0][0] === "class");
export const flattenType = (type)=>{
    const ctx = {
        type,
        path: new Path(),
        lastDomain: "undefined"
    };
    return type.meta.config ? [
        [
            "config",
            {
                config: type.meta.config,
                node: flattenNode(type.node, ctx)
            }
        ]
    ] : flattenNode(type.node, ctx);
};
export const flattenNode = (node, ctx)=>{
    if (typeof node === "string") {
        const resolution = ctx.type.meta.scope.resolve(node);
        const updatedScopeConfig = resolution.meta.scope !== ctx.type.meta.scope && resolution.meta.scope.config;
        return updatedScopeConfig ? [
            [
                "config",
                {
                    config: updatedScopeConfig,
                    node: resolution.flat
                }
            ]
        ] : resolution.flat;
    }
    const domains = keysOf(node);
    if (domains.length === 1) {
        const domain = domains[0];
        const predicate = node[domain];
        if (predicate === true) {
            return domain;
        }
        ctx.lastDomain = domain;
        const flatPredicate = flattenPredicate(predicate, ctx);
        return hasImpliedDomain(flatPredicate) ? flatPredicate : [
            [
                "domain",
                domain
            ],
            ...flatPredicate
        ];
    }
    const result = {};
    for (const domain1 of domains){
        ctx.lastDomain = domain1;
        result[domain1] = flattenPredicate(node[domain1], ctx);
    }
    return [
        [
            "domains",
            result
        ]
    ];
};
export const isLiteralNode = (node, domain)=>{
    return resolutionExtendsDomain(node, domain) && isLiteralCondition(node[domain]);
};
export const arrayOf = (node)=>({
        object: {
            class: "Array",
            props: {
                [mappedKeys.index]: node
            }
        }
    });
