var _definitions_code, _definitions_code1;
import { Scanner } from "../parse/string/shift/scanner.js";
import { domainDescriptions, domainOf } from "../utils/domains.js";
import { isWellFormedInteger } from "../utils/numericLiterals.js";
import { objectKindDescriptions } from "../utils/objectKinds.js";
import { Path } from "../utils/paths.js";
import { stringify } from "../utils/serialize.js";
export class ArkTypeError extends TypeError {
    constructor(problems){
        super(problems.summary);
        this.cause = problems;
    }
}
export class Problem {
    toString() {
        return this.message;
    }
    get message() {
        return this.writers.addContext(this.reason, this.path);
    }
    get reason() {
        return this.writers.writeReason(this.mustBe, new DataWrapper(this.data));
    }
    get mustBe() {
        return this.writers.mustBe(this.source);
    }
    constructor(code, path, data, source, writers){
        this.code = code;
        this.path = path;
        this.data = data;
        this.source = source;
        this.writers = writers;
        if (this.code === "multi") {
            this.parts = this.source;
        }
    }
}
class ProblemArray extends Array {
    add(code, data, source, opts) {
        // copy the path to avoid future mutations affecting it
        const path = opts?.path ?? Path.from(this.state.path);
        const problem = new Problem(code, path, data, source, this.state.getConfigForProblemCode(code));
        const pathKey = `${path}`;
        const existing = this.byPath[pathKey];
        if (existing) {
            if (existing.parts) {
                existing.parts.push(problem);
            } else {
                const problemIntersection = new Problem("multi", existing.path, data, [
                    existing,
                    problem
                ], this.state.getConfigForProblemCode("multi"));
                const existingIndex = this.indexOf(existing);
                // If existing is found (which it always should be unless this was externally mutated),
                // replace it with the new problem intersection. In case it isn't for whatever reason,
                // just append the intersection.
                this[existingIndex === -1 ? this.length : existingIndex] = problemIntersection;
                this.byPath[pathKey] = problemIntersection;
            }
        } else {
            this.byPath[pathKey] = problem;
            this.push(problem);
        }
        return false;
    }
    get summary() {
        return this.join("\n");
    }
    toString() {
        return this.summary;
    }
    throw() {
        throw new ArkTypeError(this);
    }
    constructor(state){
        super();
        this.state = state;
        this.byPath = {};
    }
}
export const Problems = ProblemArray;
const capitalize = (s)=>s[0].toUpperCase() + s.slice(1);
export const describeDomains = (domains)=>{
    if (domains.length === 1) {
        return domainDescriptions[domains[0]];
    }
    if (domains.length === 0) {
        return "never";
    }
    return describeBranches(domains.map((objectKind)=>domainDescriptions[objectKind]));
};
export const describeBranches = (descriptions)=>{
    let description = "";
    for(let i = 0; i < descriptions.length - 1; i++){
        description += descriptions[i];
        if (i < descriptions.length - 2) {
            description += ", ";
        }
    }
    description += ` or ${descriptions[descriptions.length - 1]}`;
    return description;
};
export class DataWrapper {
    toString() {
        return stringify(this.value);
    }
    get domain() {
        return domainOf(this.value);
    }
    get size() {
        return typeof this.value === "string" || Array.isArray(this.value) ? this.value.length : typeof this.value === "number" ? this.value : 0;
    }
    get units() {
        return typeof this.value === "string" ? "characters" : Array.isArray(this.value) ? "items" : "";
    }
    get className() {
        return Object(this.value).constructor.name;
    }
    constructor(value){
        this.value = value;
    }
}
const writeDefaultReason = (mustBe, was)=>`must be ${mustBe}${was && ` (was ${was})`}`;
const addDefaultContext = (reason, path)=>path.length === 0 ? capitalize(reason) : path.length === 1 && isWellFormedInteger(path[0]) ? `Item at index ${path[0]} ${reason}` : `${path} ${reason}`;
const compileDefaultProblemWriters = (definitions)=>{
    let code;
    for(code in definitions){
        (_definitions_code = definitions[code]).writeReason ?? (_definitions_code.writeReason = writeDefaultReason);
        (_definitions_code1 = definitions[code]).addContext ?? (_definitions_code1.addContext = addDefaultContext);
    }
    return definitions;
};
export const defaultProblemWriters = compileDefaultProblemWriters({
    divisor: {
        mustBe: (divisor)=>divisor === 1 ? `an integer` : `a multiple of ${divisor}`
    },
    class: {
        mustBe: (expected)=>typeof expected === "string" ? objectKindDescriptions[expected] : `an instance of ${expected.name}`,
        writeReason: (mustBe, data)=>writeDefaultReason(mustBe, data.className)
    },
    domain: {
        mustBe: (domain)=>domainDescriptions[domain],
        writeReason: (mustBe, data)=>writeDefaultReason(mustBe, data.domain)
    },
    domainBranches: {
        mustBe: (domains)=>describeDomains(domains),
        writeReason: (mustBe, data)=>writeDefaultReason(mustBe, data.domain)
    },
    missing: {
        mustBe: ()=>"defined",
        writeReason: (mustBe)=>writeDefaultReason(mustBe, "")
    },
    bound: {
        mustBe: (bound)=>`${Scanner.comparatorDescriptions[bound.comparator]} ${bound.limit}${bound.units ? ` ${bound.units}` : ""}`,
        writeReason: (mustBe, data)=>writeDefaultReason(mustBe, `${data.size}`)
    },
    regex: {
        mustBe: (expression)=>`a string matching ${expression}`
    },
    value: {
        mustBe: stringify
    },
    valueBranches: {
        mustBe: (values)=>describeBranches(values.map(stringify))
    },
    branches: {
        mustBe: (branchProblems)=>describeBranches(branchProblems.map((problem)=>`${problem.path} must be ${problem.parts ? describeBranches(problem.parts.map((part)=>part.mustBe)) : problem.mustBe}`)),
        writeReason: (mustBe, data)=>`${mustBe} (was ${data})`,
        addContext: (reason, path)=>path.length ? `At ${path}, ${reason}` : reason
    },
    multi: {
        mustBe: (problems)=>"• " + problems.map((_)=>_.mustBe).join("\n• "),
        writeReason: (mustBe, data)=>`${data} must be...\n${mustBe}`,
        addContext: (reason, path)=>path.length ? `At ${path}, ${reason}` : reason
    }
});
